#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#
#  title:        High Mountain Asia snow water equivalent model
#  description:  Large-scale temperature index model with several additions.
#                Model is setup for High Mountain Asia at a 0.05 degree grid.
#                This script works in conjunction with calculateSWE.cpp, which
#                is the core model, written in Rcpp/c++.
#
#  author:       Philip Kraaijenbrink <p.d.a.kraaijenbrink@uu.nl>
#  date:         23 Apr 2021
#  version:      1.0.0
#
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


# load packages ============
sapply(c('raster','stringr','tools','parallel','ncdf4',
         'abind','lubridate','Rcpp','fastmatch','caTools'),
       require, character.only=T)

# clear environment
rm(list=ls())

t1 <- Sys.time()

# predefined model extents
modelextents <- list(
  full        = c(57.375,121.125,25.875,51.875),
  full_tl     = c(57.375, 89.125, 38.875, 51.875),
  full_tr     = c(89.125, 121.125, 38.875, 51.875),
  full_bl     = c(57.375, 89.125, 25.875, 38.875),
  full_br     = c(89.125, 121.125, 25.875, 38.875),
  langtang    = c(85.125,85.975,27.975,28.625),
  centralhima = c(85.125,87.125,27.625,29.125),
  firstpixel  = c(57.625,87.875,46.425,46.825),
  annapurna   = c(84,85,28,29),
  centralhima2 = c(84.125,87.125,27.625,29.125)
)

# predefined (debug) coordinates
testcoords   <- list(
  yala         = c(85.608863,28.232285),
  yala_bc      = c(85.60867, 28.23331),
  ganja_la     = c(85.5625, 28.1545),
  ltpeak       = c(85.682705,28.381719),
  firstoffirst = c(57.65,46.45),
  paikushore   = c(85.563067,28.887956),
  rambrong     = c(84.239033, 28.4090),
  sundar       = c(84.32823, 28.417833),
  telbrung     = c(84.2794833,28.3530833)
)


# settings  ============
st <- list(

  # processing
  cores          = 6,                         # no. of logical cores to use for parallel processing when running local
  grids_in_mem   = FALSE,                     # fully load the input grids into RAM

  # paths
  outroot        = './snow-model/modelout',

  # model period
  sdate          = '1979-09-01',              # start date of model period
  edate          = '2019-08-31',              # end date of model period
  refstart       = '2000-09-01',              # start of reference period
  refend         = '2019-08-31',              # end of reference period
  tsteplen       = 1/(24/3),                  # time step length (in days)

  # model resolution and extent
  min_sdays      = 4,                         # min yearly snow cover days required for a model pixel
  modres         = 0.25/5,                    # model grid resolution, in proj units,
  modcrs         = '+init=epsg:4326',         # model grid coordinate reference system
  modext         = modelextents$full,         # model extent
  expres         = 0.5,                       # resolution at which output data is grouped in separate files
  zres           = 100,                       # elevation band equidistancy

  # temperature bias and lapse rates
  tbiasrange     = c(-30, 50),                # permissible range of optimized T biases
  lapserate      = -0.0065,                   # universal fall-back lapse rate
  lapserange     = c(-0.012,-0.002),          # permissible range of lapse rates determined from MODIS LST
  lstwindow      = 0.25,                      # size of window that is used to calculate lapserate from MODIS LST in decimal degrees

  # meteo settings
  tlow           = 0.0,                       # lower limit of liquid precip
  thigh          = 2.0,                       # upper limit of solid precip
  prcpthresh     = 0.1,                       # minimum threshold for precipitation [mm]

  # melt and storage
  ddf_low        = 4.0,                       # degree day melt factor of fresh snow [mm/K/day]
  ddf_high       = 9.0,                       # degree day melt factor of old snow [mm/K/day]
  storecap       = 0.1,                       # max water storage in snow [fraction of snow store]
  ddf_rfrz       = 1.0,                       # degree day factor for refreezing of snow melt [mm/K/day]

  # calculate components?
  calc_depth     = FALSE,                     # calculate snow compaction and also output snowdepth (Saloranta, 2016)
  calc_cover     = FALSE,                     # determine and write fractional snow cover from zband SWE calculations

  # high pass elevation band sublimation correction
  coldwinwidth   = 365,                       # with of high pass filter applied perennial swe bands (in days)
  coldprfrac     = 0.5,                       # base SWE level for perennial swe bands (in fraction of annual P)

  # elasticity runs
  do_deltas      = TRUE,                      # perform elasticity runs of combinations of T and P
  t_deltas       = seq(-1,6, 1),              # T deltas used for elasticity runs
  p_deltas       = seq(0.4,1.6,0.2),          # P deltas used for elasticity runs

  # future forcing
  do_cmip        = TRUE,                      # do runs with future deltas based on cmip6 ens means

  # output scaling
  swescale       = 10,                        # SWE values (mm) are scaled and rounded using this factor for storage
  covscale       = 10000,                     # cover values (0-1) are scaled and rounded using this factor for storage
  qscale         = 100,                       # discharge values (mm) are scaled and rounded using this factor for storage
  depthscale     = 1,                         # snow depth values (mm) are scaled and rounded using this factor for storage

  # override modulation
  pscale         = 1,                         # precipitation multiplication factor (also applied during calibration)
  tas_offset     = 0,                         # apply offset to input tas series (applied after calibration only)
  sdayscale      = 1,                         # multiplication factor for reference snow persistence, capped at 100%
  fix_tbias      = FALSE,                     # use tbias raster instead of optimization procedure, provide in indat list
  modulate_fut   = FALSE                      # apply T and P modulation also to future and CMIP runs
)


# setup time and timesteps for daily ========
dates_mod       <- seq(as.Date(st$sdate),as.Date(st$edate),'day')
dates_ref       <- seq(as.Date(st$refstart),as.Date(st$refend),'day')
tn_m            <- length(dates_mod)
tn_r            <- length(dates_ref)
months          <- 1:12
monind_mod      <- as.integer(format(dates_mod,'%m'))
monind_ref      <- as.integer(format(dates_ref,'%m'))
ymonind_mod     <- as.integer(format(dates_mod,'%Y%m'))
ymonind_ref     <- as.integer(format(dates_ref,'%Y%m'))
refmondays      <- c(table(monind_ref))
daysinmon       <- days_in_month(seq(as.Date('2018-01-01'),as.Date('2018-12-01'),'month'))

# setup time and timesteps for model time step ========
dates_mod_mts   <- rep(dates_mod, each=1/st$tsteplen)
dates_ref_mts   <- rep(dates_ref, each=1/st$tsteplen)
monind_mod_mts  <- rep(monind_mod, each=1/st$tsteplen)
monind_ref_mts  <- rep(monind_ref, each=1/st$tsteplen)
ymonind_mod_mts <- rep(ymonind_mod, each=1/st$tsteplen)
ymonind_ref_mts <- rep(ymonind_ref, each=1/st$tsteplen)

# setup five-day aggregator indices for cmip runs
daydoys         <- as.numeric(format(dates_ref,'%j'))
fivedayjulian   <- lapply(seq(1,365,5), function(x) format(seq(as.Date('2018-01-01'),as.Date('2018-12-31'),by='day')[x:(x+4)], '%m-%d'))
leapdays_ref    <- which(format(dates_ref, '%m-%d') == '02-29')
dates_noleap    <- dates_ref[-leapdays_ref]
fivedayind      <- rep(NA,length(dates_noleap))
dates_noleap_md <- format(dates_noleap,'%m-%d')
for (i in seq_along(fivedayjulian)){fivedayind[dates_noleap_md %in% fivedayjulian[[i]]] <- i}
fivedayind_ref  <- as.numeric(sprintf('%s%02d',format(dates_noleap,'%Y'),fivedayind))

# defined model extent
st$modext <- extent(st$modext)

# get elevation breaks
zbreaks    <- seq(0,9000,by=st$zres)
zbreaks[1] <- -Inf

# read input data  ========
indat_basin   <- raster('./indata/basin_mask.tif')
indat_era5_t  <- brick('./indata/era5_2m-temperature_3h_1979-2019.nc')
indat_era5_p  <- brick('./indata/era5_total-precipitation_3h_1979-2019.nc')
indat_era5_z  <- raster('./indata/era5_cellaverage_srtm.tif')
indat_z500    <- raster('./indata/srtm_elevation_500m.nc')
indat_s500    <- brick('./indata/snowpersistence_ndsi10_monclim.nc')
indat_z1000   <- raster('./indata/srtm_elevation_1000m.nc')
indat_t1000   <- brick('./indata/LST_monclim_1km.nc')
indat_snowpix <- raster('./indata/snowpersistence_ndsi10_monclim_mean.nc')
indat_tbias   <- raster('./indata/tbias_005dd.tif')

# get file names of indat
infiles <- setNames(lapply(ls(pattern='^indat*'), function(x) eval(parse(text=paste0(x,'@file@name')))),ls(pattern='^indat*'))
infiles <- data.frame(var=names(infiles), file=do.call(c,infiles), row.names=1:length(infiles))

# read cmip delta grids
cmiproot = './indata/cmip6_deltas_ensmeans'
if (st$do_cmip){
  indat_dp = lapply(list.files(cmiproot, pattern='.*_pr_.*.nc$', full.names=T),
                    function(x) readAll(brick(x)))
  indat_dt = lapply(list.files(cmiproot, pattern='.*_tas_.*.nc$', full.names=T),
                    function(x) readAll(brick(x)))
  st$cmip_exp <- str_extract(list.files(cmiproot, pattern='.*_tas_.*.nc$'), 'ssp[0-9]{3}')
}

# load grids into memory
if (st$grids_in_mem){
  for (var in ls(pattern='^indat*')){
    eval(parse(text=paste0(var,'=readAll(',var,')')))
  }
}

# get available dates from era
dates_era  <- as.Date(as.POSIXct(indat_era5_t@z[[1]], tz='UTC'))
era_refind <- dates_era %in% dates_ref
era_modind <- dates_era %in% dates_mod

# define the model grid
modgrid <- raster(crs=CRS(st$modcrs), ext=st$modext, resolution=st$modres)
# modgrid[1:length(modgrid)] <- 1:length(modgrid)

# get the snow pixels to process from modis climatology
snowpix <- as.integer(crop(indat_snowpix, st$modext) > st$min_sdays/365*10000)
snowpix <- as.integer(resample(snowpix, modgrid, method='bilinear') > 0.5)
snowpix[snowpix==0] <- NA

# get basin pixels
basinpix <- resample(crop(indat_basin,st$modext),modgrid, method='ngb')

# get cells to process
selpix  <- basinpix&snowpix
cells   <- na.omit(as.data.frame(selpix, xy=T))[,1:2]

# group cells based on defined export resolution
xbounds <- seq(st$modext[1],st$modext[2],st$expres)
ybounds <- seq(st$modext[3],st$modext[4],st$expres)
if (tail(xbounds,1) != st$modext[2]){xbounds <- c(xbounds,tail(xbounds,1)+st$expres)}
if (tail(ybounds,1) != st$modext[2]){ybounds <- c(ybounds,tail(ybounds,1)+st$expres)}
xcntr         <- (xbounds-0.5*st$expres)[-1]
ycntr         <- (ybounds-0.5*st$expres)[-1]
expcells      <- expand.grid(xcntr,ycntr)
expgrid       <- rasterFromXYZ(cbind(expcells, expcells, 1:nrow(expcells)), crs=st$modcrs)
expgrid       <- mask(resample(expgrid, selpix, method='ngb'),selpix)
expdiv        <- setNames(as.data.frame(rasterToPoints(expgrid)),c('x','y','exp_x','exp_y','exp_id'))
expdiv        <- split(expdiv[,1:4],expdiv[,5])
names(expdiv) <- sapply(expdiv, function(x) sprintf('x%06d_y%06d', as.integer(x[1,3]*1000), as.integer(x[1,4]*1000)))
expdiv        <- lapply(expdiv, function(x) matrix(unlist(x[,1:2]), nc=2))

# determine future combinations of t and t
futcombs <- expand.grid(st$t_deltas, st$p_deltas)
names(futcombs) <- c('t_delta', 'p_delta')

# load the fast SWE degree day model calculation loop in C++
sourceCpp(file='./scripts/swe-model/model/calculateSWE.cpp')

# debug cell to run
xy     <- testcoords$yala
px     <- which.min(abs(cells[,1]-xy[1]) + abs(cells[,2]-xy[2]))
coord  <- unlist(cells[px,])


# model function definition ========
sweModel <- function(coord, st=st){

  # determine the extent of the model pixel
  pxext  <- extent(c(coord[1]-st$modres/2, coord[1]+st$modres/2,
                     coord[2]-st$modres/2, coord[2]+st$modres/2))

  # get LST data within the defined window
  lapseext <- extent(c(coord[1]-st$lstwindow/2, coord[1]+st$lstwindow/2,
                       coord[2]-st$lstwindow/2, coord[2]+st$lstwindow/2))
  lstz <- values(crop(indat_z1000, lapseext, snap='out'))
  lst  <- values(crop(indat_t1000, lapseext, snap='out')) * 0.01

  # calculate monthly lapserate
  if (abs(diff(range(lstz)) > 50)){   # (only if zrange in lapsepixel is at least 50 m)
    lapsemod  <- lapply(1:12, function(m) summary(lm(lst[,m]~lstz)))
    lapserate <- sapply(lapsemod, function(x) x$coefficients[2,1])
    lapsesgnf <- sapply(lapsemod, function(x) x$coefficients[2,4])
    lapserate <- ifelse(lapsesgnf<=0.05, lapserate, st$lapserate)
    lapserate <- ifelse(lapserate < st$lapserange[1], st$lapserange[1], lapserate)
    lapserate <- ifelse(lapserate > st$lapserange[2], st$lapserange[2], lapserate)
  }else{
    lapserate <- rep(st$lapserate,12)
  }

  # get modis rasters for current modgrid pixel
  sub_z       <- values(crop(indat_z500, pxext, snap='out'))
  sub_zmean   <- mean(sub_z,na.rm=T)
  sub_scov    <- values(crop(indat_s500, pxext, snap='out')) * st$sdayscale
  sub_scov    <- ifelse(sub_scov > 10000, 10000, sub_scov)
  sub_covmean <- rowSums(t(t(sub_scov)*daysinmon / 365))   # get monthday aggregate mean

  # failsafe: only run remaining part if there are subgrid snow pixels
  if (length(na.omit(sub_scov))){

    # get elevation band data
    pixzone       <- as.numeric(cut(sub_z, breaks=zbreaks))
    bandfr        <- data.frame(zone=sort(unique(pixzone)))
    bandfr$zmid   <- bandfr$zone*st$zres - 0.5*st$zres
    bandfr$zmean  <- tapply(sub_z, pixzone, mean, na.rm=T)
    bandfr$sdays  <- tapply(sub_covmean, pixzone, mean, na.rm=T) * 0.0001 * 365
    bandfr$pixels <- tapply(rep(1,length(pixzone)), pixzone, sum, na.rm=T)
    zn            <- nrow(bandfr)

    # monthly snow cover persistence per elevation band
    monthtot     <- t(replicate(zn, refmondays))
    snowpers_obs <- do.call(cbind, lapply(1:12, function(i) tapply(sub_scov[,i], pixzone, mean, na.rm=T) * 0.0001))
    snowpers_obs <- snowpers_obs*monthtot

    # get era data
    era_z                    <- indat_era5_z[cellFromXY(indat_era5_z,coord)]
    tas                      <- c(indat_era5_t[cellFromXY(indat_era5_t,coord)]) - 273.15
    prcp                     <- c(indat_era5_p[cellFromXY(indat_era5_p,coord)]) * 1e3
    prcp[prcp<st$prcpthresh] <- 0
    tas_ref                  <- tas[era_refind]
    prcp_ref                 <- prcp[era_refind]
    tas_mod                  <- tas[era_modind]
    prcp_mod                 <- prcp[era_modind]
    pr_mod_ann               <- mean(prcp_mod)*365/st$tsteplen

    # get elevation difference per band with era's elevation
    bandfr$zlapse <- bandfr$zmean - era_z

    # get observed snow persistence for bias correction
    bpers_obs <- rowSums(snowpers_obs)

    # bias-correct ERA5 T using observed snow days
    if (st$fix_tbias){
      opt_t    = indat_tbias[cellFromXY(indat_tbias,coord)]
      opt_rmse = 0
    }else{
      optBias <- function(x){
        modpers <- sapply(1:zn, function(b){
          sweCPP(tas_ref, prcp_ref, monind_ref_mts, lapserate, zdiff=bandfr$zlapse[b],
                 bandbias=x, ddf_sm_lo=st$ddf_low, ddf_sm_hi=st$ddf_high, toffset=rep(0,12), pscale=rep(st$pscale,12), optrun=T,
                 tlimlo=st$tlow, tlimhi=st$thigh, snowsc=st$storecap, ddf_rf=st$ddf_rfrz,
                 steplen=st$tsteplen, calcdepth=F)$snowdays
        })
        rmse <- sqrt(mean((modpers-bpers_obs)^2))
        ret  <- rmse + 0.01*x^2  # add very shallow abs slope to the return function to ensure convergence
        return(ret)
      }
      opt      <- optimize(optBias, st$tbiasrange)
      opt_t    <- opt$minimum
      opt_rmse <- opt$objective
    }

    # get modelled snow days after optimization and custom offsets
    bandfr$sdays_mod <- (1/tn_r * 365) * sapply(1:zn, function(b){
      sweCPP(tas_ref, prcp_ref, monind_ref_mts, lapserate, zdiff=bandfr$zlapse[b],
             bandbias=opt_t, ddf_sm_lo=st$ddf_low, ddf_sm_hi=st$ddf_high, toffset=rep(st$tas_offset,12), pscale=rep(st$pscale,12), optrun=T,
             tlimlo=st$tlow, tlimhi=st$thigh, snowsc=st$storecap, ddf_rf=st$ddf_rfrz,
             steplen=st$tsteplen, calcdepth=F)$snowdays
      })

    # function to apply model to either model or reference period
    tasbias=opt_t; tforce=rep(st$tas_offset,12); prcpscale=rep(st$pscale,12); runtype='hist'   # debug values
    applyModel <- function(tasbias=opt_t, tforce=rep(st$tas_offset,12), prcpscale=rep(st$pscale,12), runtype='hist'){

      # get model output at zband level using bias corrected T and optimized lapse rates
      if (runtype=='hist'){
        modeldata <- lapply(1:zn, function(b){
          sweCPP(tas_mod, prcp_mod, monind_mod_mts, lapserate, zdiff=bandfr$zlapse[b],
                 bandbias=tasbias, ddf_sm_lo=st$ddf_low, ddf_sm_hi=st$ddf_high, toffset=tforce, pscale=prcpscale, optrun=F,
                 tlimlo=st$tlow, tlimhi=st$thigh, snowsc=st$storecap, ddf_rf=st$ddf_rfrz,
                 steplen=st$tsteplen, calcdepth=st$calc_depth)
          })
        ndays    <- tn_m
        timelen  <- tn_m / st$tsteplen
        aggrinds <- ymonind_mod
      }else if (runtype %in% c('fut','cmip')){
        modeldata <- lapply(1:zn, function(b){
          sweCPP(tas_ref, prcp_ref, monind_ref_mts, lapserate, zdiff=bandfr$zlapse[b],
                 bandbias=tasbias, ddf_sm_lo=st$ddf_low, ddf_sm_hi=st$ddf_high, toffset=tforce, pscale=prcpscale, optrun=F,
                 tlimlo=st$tlow, tlimhi=st$thigh, snowsc=st$storecap, ddf_rf=st$ddf_rfrz,
                 steplen=st$tsteplen, calcdepth=F)
          })
        ndays    <- tn_r
        timelen  <- tn_r / st$tsteplen
        aggrinds <- ymonind_ref
      }else{
        stop("Wrong period selection")
      }

      # extract data from model output list
      bandswe    <- do.call(cbind, lapply(modeldata, function(x) x$snowpack))
      bandq      <- do.call(cbind, lapply(modeldata, function(x) x$discharge))
      if (st$calc_depth){banddepth  <- do.call(cbind, lapply(modeldata, function(x) x$snowdepth))}
      if (st$calc_cover){bandcover  <- bandswe>0}
      relcontr   <- data.frame(snow=sapply(modeldata, function(x) x$snow_sum) / ndays*365.25,
                               rain=sapply(modeldata, function(x) x$rain_sum) / ndays*365.25)
      rosmelt    <- sapply(modeldata, function(x) x$ros_sum)  / ndays*365.25

      # filter output for bands that are 'always' below zero
      if (zn>1){
        coldbands <- colSums(bandswe > 0) / timelen > 0.95
      }else{
        coldbands <- sum(bandswe > 0) / timelen > 0.95
      }

      # For cold bands, use a proxy for daily snow loss rate equal to mean snowfall ("straightening" using a hi-pass filter)
      if (sum(coldbands)){
        bandswe[,coldbands] <- as.matrix(bandswe[,coldbands]) -
           apply(as.matrix(bandswe[,coldbands]), 2, runmean, k=st$coldwinwidth/st$tsteplen) +   # use a hi pass filter to force stable SWE
           st$coldprfrac * pr_mod_ann                                                           # set base SWE to a fraction of mean precip
        bandswe[bandswe<0]  <- 0
        if (st$calc_depth){
          banddepth[,coldbands] <- bandswe[,coldbands]/0.35                                     # use fixed density for snow in coldbands
        }
      }

      # get weighted average of outputs for the model cell
      weights   <- matrix(bandfr$pixels/sum(bandfr$pixels), nc=zn, nr=timelen, byrow=T)
      bandswe   <- bandswe * weights
      bandq     <- bandq * weights
      if (zn > 1){
        swe   <- rowSums(bandswe)
        sq    <- rowSums(bandq)
      }else{
        swe  <- bandswe
        sq   <- bandq
      }
      if (st$calc_cover){
        bandcover <- bandcover * weights
        if (zn > 1){
          scov  <- rowSums(bandcover)
        }else{
          scov <- bandcover
        }
      }
      if (st$calc_depth){
        banddepth <- banddepth * weights
        if (zn > 1){
          depth <- rowSums(banddepth)
        }else{
          depth <- banddepth
        }
      }

      # aggregate data to daily resolution
      swe   <- ctapply(swe,  rep(1:(timelen*st$tsteplen), each=1/st$tsteplen), mean)
      sq    <- ctapply(sq,   rep(1:(timelen*st$tsteplen), each=1/st$tsteplen), sum)
      if (st$calc_depth){depth <- ctapply(depth, rep(1:(timelen*st$tsteplen), each=1/st$tsteplen), mean)}
      if (st$calc_cover){scov  <- ctapply(scov,  rep(1:(timelen*st$tsteplen), each=1/st$tsteplen), mean)}

      #get monthly statistics
      mstat  <- data.frame(
        mon       = as.integer(unique(aggrinds)),
        ymon      = as.integer(substr(as.character(unique(aggrinds)),5,6)),
        swe_min   = as.integer(ctapply(swe, aggrinds, min)   * st$swescale),
        swe_mean  = as.integer(ctapply(swe, aggrinds, mean)  * st$swescale),
        swe_max   = as.integer(ctapply(swe, aggrinds, max)   * st$swescale),
        q_min     = as.integer(ctapply(sq, aggrinds, min)    * st$qscale),
        q_mean    = as.integer(ctapply(sq, aggrinds, mean)   * st$qscale),
        q_max     = as.integer(ctapply(sq, aggrinds, max)    * st$qscale),
        q_sum     = as.integer(ctapply(sq, aggrinds, sum)    * st$qscale),
        row.names = NULL)
      if (st$calc_cover){
        mstat$cov_min    = as.integer(ctapply(scov, aggrinds, min)  * st$covscale)
        mstat$cov_mean   = as.integer(ctapply(scov, aggrinds, mean) * st$covscale)
        mstat$cov_max    = as.integer(ctapply(scov, aggrinds, max)  * st$covscale)
      }
      if (st$calc_depth){
        mstat$depth_min  = as.integer(ctapply(depth, aggrinds, min)  * st$depthscale)
        mstat$depth_mean = as.integer(ctapply(depth, aggrinds, mean) * st$depthscale)
        mstat$depth_max  = as.integer(ctapply(depth, aggrinds, max)  * st$depthscale)
      }

      # return everyting for the historical run
      if (runtype=='hist'){
        histdata = list()
        histdata$swe = swe
        histdata$sq  = sq
        if (st$calc_cover){histdata$scov=scov}
        if (st$calc_depth){histdata$sdepth=depth}
        histdata$coldbands    = coldbands
        histdata$mstat        = mstat
        histdata$contribution = relcontr
        histdata$ros_melt     = rosmelt
        return(histdata)
      }

      # only return aggregate statistics of the monthly statistics for the future runs
      if (runtype=='fut'){
        climastat <- list(mean = apply(aggregate(mstat, by=list(mstat$ymon), mean)[,-c(1,2)],c(2),as.integer),
                          sd   = apply(aggregate(mstat, by=list(mstat$ymon), sd)[,-c(1,2)],c(2),as.integer))
        return(climastat)
      }

      # return aggregate statistics as well as 5-day aggregates series of swe and Q for the CMIP runs
      if (runtype=='cmip'){
        climastat <- list(mean = apply(aggregate(mstat, by=list(mstat$ymon), mean)[,-c(1,2)],c(2),as.integer),
                          sd   = apply(aggregate(mstat, by=list(mstat$ymon), sd)[,-c(1,2)],c(2),as.integer))
        fiveday <- list(swe = ctapply(swe[-leapdays_ref], fivedayind_ref, mean),
                        q   = ctapply(sq[-leapdays_ref],  fivedayind_ref, mean))
        return(list(climastat=climastat, fiveday=fiveday))
      }

    }

    # run model for historical period
    historical      <- applyModel(tasbias=opt_t,
                                  tforce=rep(st$tas_offset,12),
                                  prcpscale=rep(st$pscale,12),
                                  runtype='hist')
    bandfr$coldband <- historical$coldbands
    bandfr$snow     <- historical$contribution$snow
    bandfr$rain     <- historical$contribution$rain
    bandfr$ros_melt <- historical$ros_melt

    # prepare list of output data per cell
    outdata <- list(
      zbands      = bandfr,
      monlapse    = lapserate,
      tbias       = opt_t,
      rmse        = sqrt(mean((bandfr$sdays_mod - bandfr$sdays)^2)),
      coord       = coord,
      histstat    = historical$mstat,
      swe         = historical$swe,
      q           = historical$sq
    )
    if (st$calc_cover){outdata$cover = historical$scov}
    if (st$calc_depth){outdata$depth = historical$sdepth}

    # set T and P offset for future conditions and projections
    if (st$modulate_fut){
      fut_toffset = rep(st$tas_offset, 12)
      fut_pscale  = rep(st$pscale, 12)
    }else{
      fut_toffset = rep(0,12)
      fut_pscale  = rep(1,12)
    }

    # run model for set of synthetic conditions
    if (st$do_deltas){
      future <- lapply(1:nrow(futcombs), function(comb){
        applyModel(tasbias=opt_t,
                   tforce=rep(futcombs$t_delta[comb],12)+fut_toffset,
                   prcpscale=rep(futcombs$p_delta[comb],12)*fut_pscale,
                   runtype='fut')})
      outdata$futstat <- list(mean = do.call(abind,list(lapply(future,function(x) x$mean),along=3)),
                              sd   = do.call(abind,list(lapply(future,function(x) x$sd),along=3)))
    }

    # run model for set of future projections
    if (st$do_cmip){
      delta_p <- lapply(indat_dp, function(x) c(x[cellFromXY(x,coord)]))
      delta_t <- lapply(indat_dt, function(x) c(x[cellFromXY(x,coord)]))
      cmiprun <- lapply(1:length(delta_p), function(i_rcp){
        applyModel(tasbias   = opt_t,
                   tforce    = delta_t[[i_rcp]],
                   prcpscale = delta_p[[i_rcp]],
                   runtype   = 'cmip')})
      names(cmiprun) <- st$cmip_exp
      outdata$cmip   <- cmiprun
    }

    # clean up
    rm(list=setdiff(ls(), c("outdata")))
    invisible(gc())

    return(outdata)
  }
}




# run model ============================================

# check dirs in output directory
outroot   <- file.path(st$outroot)
modroot   <- file.path(st$outroot,'modout')
metaroot  <- file.path(st$outroot,'meta')
dir.create(modroot, showWarnings=F, recursive=T)
dir.create(metaroot, showWarnings=F, recursive=T)
fnpresent <- list.files(modroot)
missing   <- which(!(names(expdiv) %in% fnpresent))
exp2run   <- missing

# write metadata to disk
saveRDS(st, file=file.path(metaroot,'settings.rds'))
saveRDS(list(dates_mod=dates_mod, dates_ref=dates_ref), file=file.path(metaroot,'dates.rds'))
write.csv(infiles, file.path(metaroot,'input_files.csv'), row.names=F)
write.csv(futcombs, file.path(metaroot, 'delta_combinations.csv'), row.names=F)

# model i/o controller
modelController <- function(pixelframe, outdirname, st=st){

  outdir   <- file.path(modroot, outdirname)
  dir.create(outdir, showWarnings=F, recursive=T)
  exp_px   <- split(pixelframe, seq(nrow(pixelframe)))
  modelout <- lapply(exp_px, sweModel, st=st)
  modelout <- modelout[!sapply(modelout, is.null)]

  # extract separate variables
  coorddat <- do.call(rbind,lapply(modelout, function(x) x$coord))
  zbanddat <- lapply(modelout, function(x) x$zbands)
  lapsedat <- do.call(rbind,lapply(modelout, function(x) x$monlapse))
  biasdat  <- do.call(c,lapply(modelout, function(x) x$tbias))
  rmsedat  <- do.call(c,lapply(modelout, function(x) x$rmse))
  swedat   <- do.call(cbind,lapply(modelout, function(x) x$swe))
  qdat     <- do.call(cbind,lapply(modelout, function(x) x$q))
  hstatdat <- do.call(abind, list(lapply(modelout, function(x) x$histstat), along=3))
  if (st$calc_depth){depthdat <- do.call(cbind,lapply(modelout, function(x) x$depth))}
  if (st$calc_cover){coverdat <- do.call(cbind,lapply(modelout, function(x) x$cover))}

  # scale swe and cover data for storage and improved io
  swedat   <- apply(swedat, c(2), function(x) as.integer(x*st$swescale))
  qdat     <- apply(qdat, c(2), function(x) as.integer(x*st$qscale))
  if (st$calc_depth){depthdat <- apply(depthdat ,c(2), function(x) as.integer(x*st$depthscale))}
  if (st$calc_cover){coverdat <- apply(coverdat ,c(2), function(x) as.integer(x*st$covscale))}

  # save as rds
  saveRDS(coorddat, file.path(outdir,'coord.rds'))
  saveRDS(zbanddat, file.path(outdir,'zbands.rds'))
  saveRDS(lapsedat, file.path(outdir,'lapse.rds'))
  saveRDS(biasdat,  file.path(outdir,'bias.rds'))
  saveRDS(rmsedat,  file.path(outdir,'rmse.rds'))
  saveRDS(swedat,   file.path(outdir,'swe.rds'))
  saveRDS(qdat,     file.path(outdir,'q.rds'))
  saveRDS(hstatdat, file.path(outdir,'histmonstat.rds'))
  if (st$calc_depth){saveRDS(depthdat, file.path(outdir,'depth.rds'))}
  if (st$calc_cover){saveRDS(coverdat, file.path(outdir,'cover.rds'))}

  # save future runs if selected
  if (st$do_deltas){
    fstatdat <- lapply(modelout, function(x) x$futstat)
    saveRDS(fstatdat, file.path(outdir,'futstat.rds'))
  }

  # save cmip runs if selected
  if (st$do_cmip){
    cmipdat <- lapply(modelout, function(x) x$cmip)
    saveRDS(cmipdat, file.path(outdir,'cmip.rds'))
  }
}

# record init time
t2 <- Sys.time()
tdiff_init <- round(difftime(t2,t1,units='mins'),2)
cat(sprintf('\nInitialization time is %.2f minutes\n',as.numeric(tdiff_init)))

# run multithreaded
cl = makeCluster(st$cores,'PSOCK')
varlist = c(
  'st',ls(patt='^indat_.*'),'dates_mod','dates_ref','tn_m','tn_r','months',
  'monind_mod_mts','monind_ref_mts','fivedayind_ref','leapdays_ref',
  'monind_mod','monind_ref','refmondays','daysinmon','zbreaks','futcombs',
  'infiles','dates_era','era_refind','era_modind','ymonind_mod','ymonind_ref',
  'modelController','sweModel','expdiv','outroot','modroot','metaroot'
)
clusterExport(cl, varlist)
clusterEvalQ(cl, sapply(c('abind','raster','Rcpp','fastmatch','caTools'), require, character.only=T))
clusterEvalQ(cl, sourceCpp(file='./scripts/swe-model/model/calculateSWE.cpp'))
out <- parSapply(cl, exp2run, function(i) modelController(expdiv[[i]], names(expdiv)[i], st=st))
stopCluster(cl)


# print elapsed time
t3 <- Sys.time()
tdiff_par  <- round(difftime(t3,t2,units='mins'),2)
tdiff_tot  <- round(difftime(t3,t1,units='mins'),2)
cat(sprintf('Parallel time is %.2f minutes\n',as.numeric(tdiff_par)))
cat(sprintf('Total time is %.2f minutes\n\n',as.numeric(tdiff_tot)))
